% The ALE Library
%
% created: 4 May, 2006 by Gerald Penn

:- multifile if/2, macro/2, fun/1, fsmode/1.
:- discontiguous if/2, macro/2, fun/1, fsmode/1.

% for the denotationally inclined...
fun top(-).
top(_) if fail.

false if fail.

% a 2-place inequation relation
ineq(X,(=\= X)) if true.

% a delayed type negation macro - wait until an FS reaches type T, then die.
not(T) macro not(a_ T).
fun not(+,-).
not((a_ T),FS) if
  prolog(ale(when(FS=T,fail))).  % this sleight of hand prevents the
   % compiler from compiling the when/2 call (which would treat T as a
   % wide-scope variable because it isn't bound yet) until not/2
   % is actually called.

fsmode append(-sub(list),-list,-list).
append([],(L,list),L) if true.
append([H|L1],L2,[H|L3]) if
  append(L1,L2,L3).

% concat/3 is one of the popular modes of append/3
fsmode concat(-sub(list),-list,-list).
concat(L1,L2,L3) if
  when(L1=sub(list),concat_(L1,L2,L3)).

fsmode concat_(+sub(list),-list,-list).
concat_([],(L,list),L) if true.
concat_([H|L1],L2,[H|L3]) if
  concat(L1,L2,L3).

% L1*L2 concatenates L1 and L2
fun '*'(+,+,-).
fsmode '*'(-sub(list),-list,-list).
'*'(L1,L2,L3) if concat(L1,L2,L3).

% split/3 is the other popular mode of append/3
fsmode split(-sub(list),-sub(list),-list).
split(L3,L1,L2) if
  when(L3=sub(list),split_(L3,L1,L2)).

fsmode split_(+sub(list),-sub(list),-list).
split_(L,[],L) if true.
split_([H|L3],[H|L1],L2) if
  split(L3,L1,L2).

% member/2 formulated to avoid choicepoint on last elt.
fsmode member(bot,-ne_list).
member(Elt,[Head|Tail]) if
  member_(Tail,Head,Elt).

fsmode member_(+list,bot,bot).
member_(_,Elt,Elt) if true.
member_([Head|Tail],_,Elt) if
  member_(Tail,Head,Elt).

% memberchk/2 refuses to backtrack on its 1st arg.
fsmode memberchk(bot,-ne_list).
memberchk(Elt,[Elt|_]) if !.
memberchk(Elt,[_|Rest]) if
  memberchk(Elt,Rest).

% find/2 is like member/2, but won't generate a list
fsmode find(bot,-ne_list).
find(Elt,List) if
  when(List=sub(list),find_act(List,Elt)).

fsmode find_act(sub(list)-ne_list,bot).
find_act([Head|Tail],Elt) if
  find_(Tail,Head,Elt).

fsmode find_(+list,bot,bot).
find_(_,Elt,Elt) if true.
find_(List,_,Elt) if
  when(List=sub(list),find_act(List,Elt)).

% findonce/2 is like memberchk/2, but refuses to backtrack on its 1st arg.
fsmode findonce(bot,-ne_list).
findonce(Elt,List) if
  when(List=sub(list),findonce_act(List,Elt)).

fsmode findonce_act(sub(list)-ne_list,bot).
findonce_act([Elt|_],Elt) if !.
findonce_act([_|Rest],Elt) if
  findonce(Elt,Rest).
