% $Id: save_results.pl,v 1.15 2003/07/14 18:29:00 dm Exp $
%
% This file defines a pretty printing hook that
% saves what is being printed for future use
% (like printing again, or for a diff, etc.)
%
% This hook does not actually print anything. Instead, it fails 
% after recording the feature structure to be printed and
% thereby lets other hooks or the default pretty printer create
% the actual output that was requested.
%
% Note: This file must be loaded before any other files
% defining portray_fs/10 printing hooks in order to have an effect!
%
% Author: Detmar Meurers
% Modified by: Gerald Penn
% ------------------------------------------------------------------

:- dynamic saved_result/2,saved_id/1,save_results_flag/1.

% --------
% switches:

save_results_on :-
   (save_results_flag
   ->  format(user_error,'~N**Error: Results are already being saved.~n* You can switch it off with save_results_off/0.~n',[] )
   ;   (  (saved_id(N), N>0)
       -> query_user('~NDo you want to keep the ~w results you already saved? (return for <~w>) ',[N,no],no,Answer),
	   ( ( Answer == n
	     ; Answer == 'n.'
	     ; Answer == no
	     ; Answer == 'no.' ) 
	   ->  retractall(saved_result(_,_)),
	       retractall(saved_id(_)),
	       assert(saved_id(0))
	   ;   true
	   )
       ;   retractall(saved_id(_)),
           asserta(saved_id(0))
       ),
       asserta(save_results_flag(true))
   ).

save_results_off :-
   (save_results_flag
   ->  saved_id(Nr),
       format('~NSaved a total of ~p results.~n',[Nr]),
       quiet_save_results_off
   ;   format(user_error,'~N**Error: Saving of results is not switched on.~n  You can enable it with save_results_on/0.~n',[])
   ).

quiet_save_results_off :-
   asserta(save_results_flag(fail)).

save_results_restore :-    % restore to previous value
   retract(save_results_flag(_)).

save_results_flag :-	     % 0-arity pred for asking the value of this flag
   save_results_flag(X), !,
   call(X).

% ------------------------------------------------------------------
:- multifile portray_fs/11.

% GBP - ported; this doesn't appear to print anything - it's just a hook for saving results.
% Unclear whether it was meant to be called before or after the real GRALE portray_fs/11.
portray_fs(_,FS,_,_,Vis0,_,_,_,_,_,_):-
   save_results_flag,		% fails immediately if flag not on
   (empty_assoc(Vis0)		% nothing visited -> top level
   ->  retract(saved_id(ID)),
       NewID is ID + 1,
       assert(saved_id(NewID)),
       asserta(saved_result(NewID,FS)),
       format('~NSaved output as result ~p.~n~n',[NewID])
   ;   true),
   fail.	% predicate always fails to get the pretty printer going
   

% GBP - ported, although it looks like residues are being stripped off.
% show_saved/1 does not deal with potential residues. Code in the if
% branch should probably be replaced with a call to mgs/1 if we can
% stop that from producing additional user output.
show_saved(Nr) :-
   (saved_result(Nr,FS) 
   ->  quiet_save_results_off,	% don't save again what we've already got
       ((current_predicate(portray_cat,portray_cat(_,_,_,_)),
	 portray_cat(['{saved',result,Nr,'}'],bot,FS,[])) -> true
				% GBP - Nr was formerly the first argument;
				% this violates the portray_cat/4 input syntax.
	 ; pp_fs(FS)
	 ),
       save_results_restore	% switch back to previous value
   ;   format(user_error,'~N**Error: No saved feature structure number ~p ',[Nr]),
       (saved_id(Total) 
       ->  format(user_error,'(only ~p were saved).~n',[Total])
       ;   format(user_error,'(none were saved).~n',[])
       )
   ).     


show_all_saved :-
   show_saved(_),
   fail.
show_all_saved.

save_results(Filename) :-
   (saved_result(Nr,_) 
   ->  format('~NSaving ~p results to file ~p ... ~n',[Nr,Filename]),
       tell(Filename),
       saved_id(Id),
       format(':- assert(saved_id(~p)).~n',[Id]),
       save_results_act,
       told
   ;   format(user_error,'~N**Error: There are no saved results to be written to a file.~n',[])
   ).     

% GBP - this note is probably wrong.  This fs2desc/2 call also strips off
%  inequations, which is lame.
% Note: The call to fs2desc/2 also modifies FS, so it's important this is
% a failure driven loop (i.e. we always fail after calling fs2desc/2).
save_results_act :-
   saved_result(Nr,FS),
   fs2desc(FS,Desc),
   portray_clause(saved_result_desc(Nr,(Desc))),
   fail.
save_results_act.

load_results(Filename) :-
   (  (saved_id(N), 
       N>0, 
       query_user('Some saved results are present, which loading results from the file will erase. Proceed anyways? (return for <~w>) ',[no],no,Answer),
       ( Answer == n
       ; Answer == 'n.'
       ; Answer == no
       ; Answer == 'no.' )) 
   ->  format('~Nload_results/1 aborted~n',[])
   ;   format('~NLoading results from file ~p~n.',[Filename]),
       retractall(saved_id(_)),
       retractall(saved_result(_,_)),
       [Filename],
       secret_noadderrs,
       load_results_act,
       secret_adderrs
   ).

% GBP - ported
load_results_act :-
   (  current_predicate(saved_result_desc,saved_result_desc(_,_))
   -> saved_result_desc(Nr,Desc),
      ( secret_noadderrs
      ; secret_adderrs,
	fail
      ),
      call_residue(add_to(Desc,NewFS),_Residue),
      asserta(saved_result(Nr,NewFS)),
      fail
   ;  format(user_error,'~N**Warning: The file contained no Trale data packages.~n',[])
   ).
load_results_act :-
   secret_adderrs.
   
