%%% ==================================================================
%%% Main startup file to load the TRALE system
%%%
%%% This file expects to be loaded in a shell in which certain
%%% system variables are set, in particular TRALE_HOME, which should
%%% point to the directory including the trale prolog files. The normal
%%% way of starting the system therefore is by executing the "trale" 
%%% shell script, which then loads this file.
%%%
%%% ==================================================================

% Define version of this release:

trale_milca_version('2.7.12').

% ----------------------------------------------------------------------------
% Load sicstus support code
% GBP - this imports environ/2 from library(system).
:- ensure_loaded(sics).

% ----------------------------------------------------------------------------
% GBP - bad move; SICStus can't resolve an embedded ~ like this on
%  some platforms.  Moved this default to the trale shell script,
%  which we read back in through TRALE_USER_STARTUP.
% user_startup_file('~/.trale_startup.pl').
:- multifile user_startup_file/1.
:- dynamic user_startup_file/1.
:- ( environ('TRALE_USER_STARTUP',F) -> assertz(user_startup_file(F))
   ; true).

% ----------------------------------------------------------------------------
% Handle user startup file

%% Load user startup file, if exists and not prohibited
load_user_startup :-
   user_startup_file(F0),
   absolute_file_name(F0,F),
   ( file_exists(F), environ('READ_USER_STARTUP',true) 
   ->  format('~NLoading user startup file ~w~n',[F0]),
       load_files(F,[if(changed)]) 
   ;   true
   ).

:- load_user_startup.

% ----------------------------------------------------------------------------
% Define trale_home file search path, if not yet defined, on the basis of 
% the TRALE_HOME environment variable, which has to be set in the shell 
% before this code is loaded. (Note that we here don't make use of 
% trale_search_path_add/2 since it's not yet defined at this point, but
% only in trale.pl, which we want to keep loadable by itself).

:- multifile file_search_path/2.  % GBP - need to declare this or else your
                                  %  search paths will be over-written when
                                  %  someone else adds one.
:- dynamic file_search_path/2.    % GBP - this avoids having to check with
                                  %  predicate_property/2 first (which should
                                  %  have been current_predicate/2 anyway -
                                  %  predicate_property/2 backtracks).
	
:- ( environ('TRALE_HOME',Path) 
   -> absolute_file_name(Path,AbsPath),  % don't load this her
%      predicate_property(user:file_search_path(_,_),_),
      (  user:file_search_path(trale_home,AbsPath) -> true
      ;  assertz(user:file_search_path(trale_home,AbsPath))
      )
   ;  format(user_error,'**Error: Set variable TRALE_HOME to directory housing the trale code.~n',[]),
%      halt ).  % GBP - halt kills the process, which may take the error msg with it
      abort ).
       
% ----------------------------------------------------------------------------
% Load the various parts of the TRALE system
% ----------------------------------------------------------------------------

% core system:
:- ensure_loaded(trale_home('trale.pl')).

% additional libraries:
% :- ensure_loaded(trale_home(lib)). % GBP: bad idea - we put it into a library
                                     %  so that people didn't have to use it.

% debugger code:
% :- ensure_loaded(trale_home(interp)).  % GBP - good idea, but the debugger isn't
% :- ensure_loaded(trale_home(trinterp)).  % ready for this port yet.

% Load the OSU extensions
:- nl,write('
OSU TRALE Modules 1.0
Copyright (C) 2002, Detmar Meurers
Copyright (C) 2006, Gerald Penn	   
All rights reserved'),nl,nl.

% code for mapping from feature structures to ALE descriptions
:- ensure_loaded(trale_home(fs2desc)).      

% predicates allowing types like ne_list, e_list and parser related 
% features like dtrs and phon to be user-defined
:- ensure_loaded(trale_home(grammar_interface)).    

% feature ordering support (currently only effective with grisu interface):
:- ensure_loaded(trale_home(feature_ordering)).    

% utilities for working with macros hiearchies
:- ensure_loaded(trale_home(macro_utils)). 

% facilty for generation of access macros
:- ensure_loaded(trale_home(extend_access)).    

% test suite support:
:- ensure_loaded(trale_home(test_suite_handling)). 

% some other extras (like saving states etc.)
:- ensure_loaded(trale_home(trale_extras)).        

% GBP - maybe once I figure out what it actually does...
% lexical rule compiler
%:- ensure_loaded(trale_home('lr_compiler/main')).        

% GBP - probably will not work with -fqs 
% graphviz hierarchy display
:- ensure_loaded(trale_home('graphviz')).

% GBP - probably will not work with -fqs
% chart display support code
:- ensure_loaded(trale_home('chart_display/chart_display')).

% support for saving results that are printed, e.g.
% for reprinting them (e.g. with another pretty printer),
% saving them to a file, or doing a diff on results
% Note: This needs to be loaded before any other
% pretty printing hooks in order to have any effect.
:- ensure_loaded(trale_home(save_results)).    

% load code for the graphical interface, if needed
:- (   environ('INTERFACE',text) ->  true
   ;   % load pretty printing hooks for grisu interface:
       ensure_loaded(trale_home(ghooks)),
       % diff of two feature structures
       ensure_loaded(trale_home(diff)),
       % debug unifies two feature structures
       ensure_loaded(trale_home(debug_unify))   
   ).

% tsdb support: not (yet) generally loaded, call load_tsdb/0 to load
load_tsdb :- ensure_loaded(trale_home('tsdb/itsdb')),
             ensure_loaded(trale_home('tsdb/trale-itsdb')).

% GBP - I don't even have the .el code for this
% load code that is called from the EMACS menu
% StMue - it is now in the SVN
:- ensure_loaded(trale_home('trale_emacs')).

% call the startup code
:- trale_startup.

